/*

*/

import tokenizer.*;

/**
   Expression Language E11 (compare with E3).

      E ::= E '+' T    // left recursion means left associative
          | E '-' T    // left recursion means left associative
          | T

      T ::= T '*' F    // left recursion means left associative
          | T '/' F    // left recursion means left associative
          | F

      F ::= '-' F      // right recursion means right associative
          | '+' F      // right recursion means right associative
          | '(' E ')'
          | 'a'

   We can rewrite the grammar this way.

      E ::= T ( ( '+' | '-' ) T )*   // sum of Terms (iteration)
      T ::= F ( ( '*' | '/' ) F )*   // product of Factors (iteration)
      F ::= ( '+' | '-' ) F
          | '(' E ')'
          | 'a'

   This parser does not "do" anything.
   It does not create a parse tree nor
   an abstract syntax tree. This parser
   is just a "recognizer" for the language.
   It will throw an error if the string cannot
   be parsed.

   Some example strings are:
      "-a"
      "+a"
      "+(-a)"
      "-(+a)"
      "+-(-+a)"
      "(+a + -a)"
      "-(a + a) / -a"
      "-a * -(a + a)"
      "(a + a) * -(a / a)"
      "-a / (a + -(a * -a))"
      "-(-(-a + -a) * -a) + -a"
   This grammar can handle any valid four function
   arithmetic expression.
*/
public class Parse_E11v2
{
   public static void recognizeE11(String stringInE11)
   {
      // Tokenize the input string.
      final Tokenizer tokens = new Tokenizer(stringInE11);
      // Parse the list of tokens.
      getE(tokens);
      // There shouldn't be any remaining tokens.
      if ( ! tokens.isEmpty() )
      {
         throw new ParseError("unexpected tokens, found " + tokens.peek());
      }
   }


   /**
      E ::= T ( ( '+' | '-' ) T )*
   */
   private static void getE(Tokenizer tokens)
   {
      getT(tokens);
      while ( ( ! tokens.isEmpty()) && ( tokens.match("+")
                                      || tokens.match("-") ) )
      {
         getT(tokens);
      }
   }


   /**
      T ::= F ( ( '*' | '/' ) F )*
   */
   private static void getT(Tokenizer tokens)
   {
      getF(tokens);
      while ( ( ! tokens.isEmpty()) && ( tokens.match("*")
                                      || tokens.match("/") ) )
      {
         getF(tokens);
      }
   }


   /**
      F ::= ( '+' | '-' ) F
          | '(' E ')'
          | 'a'
   */
   private static void getF(Tokenizer tokens)
   {
      if ( tokens.match("-")
        || tokens.match("+") )
      {
         getF(tokens);
      }
      else if ( tokens.match("(") )
      {
         getE(tokens);
         if ( ! tokens.match(")") )
         {
            throw new ParseError("expected \")\", found " + tokens.peek());
         }
      }
      else
      {
         if ( ! tokens.match("a") )
         {
            throw new ParseError("expected \"a\", found " + tokens.peek());
         }
      }
   }


   public static void main(String[] args)
   {
      String[] expressions = {
                              "-a",
                              "+a",
                              "+(-a)",
                              "-(+a)",
                              "+-(-+a)",
                              "(+a + -a)",
                              "-(a + a) / -a",
                              "-a * -(a + a)",
                              "(a + a) * -(a / a)",
                              "-a / (a + -(a * -a))",
                              "-(-(-a + -a) * -a) + -a",
                              "-a * -(-a + -a) / +a",
                              "-a / -(+(-a - -a)) * -a",
                              "a - (a * a) + a",
                              "a * (a + a) - a",
                              "a * (a + a) - -a",
                              "-(a + a + a) - a",
                              "a + -(a * a)) + a",
                              "a + -(-(a * a) + a"
                             };

      for (final String str : expressions)
      {
         try
         {
            System.out.print( "Parsing: \"" + str + "\"" );
            recognizeE11( str );
            System.out.println( " ... Parsed" );
         }
         catch (TokenizeError e)
         {
            System.out.println( " ... Failed (" + e + ")" );
         }
         catch (ParseError e)
         {
            System.out.println( " ... Failed (" + e + ")" );
         }
      }
   }
}
