/*

*/

import tokenizer.*;

/**
   Expression Language E3.

      E ::= E '+' 'a'   // left recursion means left associative
          | 'a'

   We cannot parse this language, as it is written, using a
   recursive descent parser. The left recursion in the grammar
   causes infinite recursion in the recursive descent parser.
   We must rewrite the grammar to remove the left recursion
   and replace the recursion with a kind of iteration.

   We want to show that the above grammar can be rewritten to
   use the Kleene star (from "regular expressions") in place
   of the left recursion. The new grammar needs only one
   production.

      E ::= 'a' ( '+' 'a' )*    // * means zero-or-more

   Let us look at how the left recursion is removed from
   the original grammar to derive the new grammar.

   Here is a sequence of left-most derivations of sentential
   forms from the original grammar.

                      E
                  E + a
              E + a + a
          E + a + a + a
      E + a + a + a + a
      a + a + a + a + a

   Notice how the "expression part" is moving to the left, and in
   each step it grows the expression by concatenating the string
   "a +" onto what had been to its right. So that means we can think
   of growing a string in the language by starting with the sting "a"
   and then concatenating on the left as many "a +" strings as we like.
   This leads to the following production for the language (which is
   not yet the production that we want)

       E ::= ( 'a' '+' )* 'a'

   which uses iteration (the Kleene star) in place of recursion.

   But we can just as reasonably look at the final string
      a + a + a + a + a
   and say that we grow this string by starting with the string "a"
   and then concatenating on the right(!) as many "+ a" strings as
   we like. Hence, we can also derive this production for the
   language, which is the production that we want.

       E ::= 'a' ( '+' 'a' )*

   So the left recursion in this production

       E ::=  E '+' 'a' | 'a'

   can be factored out to derive this non-recursive production.

       E ::= 'a' ( '+' 'a' )*

   It is easy to translate this production into a while-loop that
   parses the language. But there is still a problem. Let us go
   back to the right recursive expression language E2,

      E2 ::= 'a' '+' E2   // right recursion means right associative
           | 'a'

   and use the same trick to factor out the recursion.

   Here is a sequence of left-most derivations of sentential
   forms from the E2 grammar.

      E2
      a + E2
      a + a + E2
      a + a + a + E2
      a + a + a + a + E2
      a + a + a + a + a

   Notice how the "expression part" is moving to the right, and in
   each step it grows the expression by concatenating the string
   "+ a" onto what had been to its left. So that means we can think
   of growing a string in the E2 language by starting with the sting
   "a" and then concatenating on the right as many "+ a" strings as
   we like. This leads to the following production for language E2.

       E2 ::= 'a' ( '+' 'a' )*

   But this is exactly the same production we derived for language E3!
   Language E2 is supposed to be right associative and language E3 is
   supposed to be left associative, but it they have the same grammar,
   how can we say which associativity they have?

   The answer is that we cannot. This production

       E ::= 'a' ( '+' 'a' )*

   does not define an associativity for the '+' operator. Instead, the
   associativity of the '+' operator will be determined by how we write
   the parser, not by how we write the production. When the parser code
   uses this production to parse an expression, we can have the parser
   code build the parse tree as either a left associative or a right
   associative parse tree.

   Here is the non-recursive pseudo code that parses the iterative
   production.

      getE(tokens)
      {
         tokens.match("a");
         while ( ! tokens.isEmpty() && tokens.match("+") )
         {
            tokens.match("a");
         }
      }

   This parser does not "do" anything. It does not create
   a parse tree nor an abstract syntax tree. This parser
   is just a "recognizer" for the language. It will throw
   an error if the string cannot be parsed.

   Some example strings are:
      "a"
      "a + a"
      "a + a + a"
      "a + a + a + a"
*/
public class Parse_E3
{
   public static void recognizeE3(String stringInE3)
   {
      // Tokenize the input string.
      final Tokenizer tokens = new Tokenizer(stringInE3);
      // Parse the list of tokens.
      getE(tokens);
      // There shouldn't be any remaining tokens.
      if ( ! tokens.isEmpty() )
      {
         throw new ParseError("unexpected tokens, found " + tokens.peek());
      }
   }


   /**
      E ::= 'a' ( '+' 'a' )*
   */
   private static void getE(Tokenizer tokens)
   {
      if ( ! tokens.match("a") ) // consumes one token
      {
         throw new ParseError("expected \"a\", found " + tokens.peek());
      }
      while ( ( ! tokens.isEmpty()) && tokens.match("+") )
      {
         if ( ! tokens.match("a") )
         {
            throw new ParseError("expected \"a\", found " + tokens.peek());
         }
      }
   }

/*
   // Here is a simplified, partial parser. It always
   // succeeds on a syntactically correct input string,
   // but it can give incorrect results for syntactically
   // incorrect input strings (even go into an infinite
   // loop). Notice that this is just the "pseudo code"
   // parser that we wrote above.
   private static void getE(Tokenizer tokens)
   {
      tokens.match("a");
      while ( ! tokens.isEmpty() && tokens.match("+") )
      {
         tokens.match("a");
      }
   }
*/

/*
   // This is what the left recursive grammar would
   // lead us to write if we tried to write a recursive
   // descent parser for the grammar without any changes.
   private static void getE(Tokenizer tokens)
   {
      if ( tokens.size() > 1 )
      {
         getE(tokens); // recursion
         if ( ! tokens.match("+") )
         {
            throw new ParseError("expected \"+\"");
         }
         if ( ! tokens.match("a") )
         {
            throw new ParseError("expected \"a\"");
         }
      }
      else
      {
         if ( ! tokens.match("a") )
         {
            throw new ParseError("expected \"a\"");
         }
      }
   }
*/

   public static void main(String[] args)
   {
      String[] expressions = {
                              "a",
                              "a + a",
                              "a + a + a",
                              "a + a + a + a",
                              "a+a+a+a",
                              "a+a+a-a",
                              "a + a +",
                              "+ a + a"
                             };

      for (final String str : expressions)
      {
         try
         {
            System.out.print( "Parsing: \"" + str + "\"" );
            recognizeE3( str );
            System.out.println( " ... Parsed" );
         }
         catch (TokenizeError e)
         {
            System.out.println( " ... Failed (" + e + ")" );
         }
         catch (ParseError e)
         {
            System.out.println( " ... Failed (" + e + ")" );
         }
      }
   }
}
