/*

*/

import tokenizer.*;

/**
   Expression Language E7 (with right recursion).

      E ::= 'a' '+' E   // right recursion means right associative
          | 'a' '-' E   // right recursion means right associative
          | 'a'

   We can rewrite the grammar this way.

      E ::= 'a' [ ( '+' | '-' ) E ]   // right recursion means right associative

   This production says that an E expands into an
   'a' followed by an optional plus-or-minus E

   This parser does not "do" anything.
   It does not create a parse tree nor
   an abstract syntax tree. This parser
   is just a "recognizer" for the language.
   It will throw an error if the string cannot
   be parsed.

   Some example strings are:
      "a"
      "a + a"
      "a - a"
      "a - a + a"
      "a + a - a - a"
*/
public class Parse_E7
{
   public static void recognizeE7(String stringInE7)
   {
      // Tokenize the input string.
      final Tokenizer tokens = new Tokenizer(stringInE7);
      // Parse the list of tokens.
      getE(tokens);
      // There shouldn't be any remaining tokens.
      if ( ! tokens.isEmpty() )
      {
         throw new ParseError("unexpected tokens, found " + tokens.peek());
      }
   }


   /**
      E ::= 'a' [ ( '+' | '-' ) E ]
   */
   private static void getE(Tokenizer tokens)
   {
      if ( ! tokens.match("a") )
      {
         throw new ParseError("expected \"a\", found " + tokens.peek());
      }
      if ( ! tokens.isEmpty() && ( tokens.match("+")
                                || tokens.match("-") ) )
      {
         getE(tokens); // recursion
      }
   }


   public static void main(String[] args)
   {
      String[] expressions = {
                              "a",
                              "a + a",
                              "a - a",
                              "a - a + a",
                              "a + a - a - a",
                              "a-a-a+a",
                              "a+a+a-a",
                              "a + + a + a",
                              "a + a a + a",
                              "a - a +",
                              "- a + a",
                              "+ a - a"
                             };

      for (final String str : expressions)
      {
         try
         {
            System.out.print( "Parsing: \"" + str + "\"" );
            recognizeE7( str );
            System.out.println( " ... Parsed" );
         }
         catch (TokenizeError e)
         {
            System.out.println( " ... Failed (" + e + ")" );
         }
         catch (ParseError e)
         {
            System.out.println( " ... Failed (" + e + ")" );
         }
      }
   }
}
