/*

*/

import tokenizer.*;


/**
   Expression Language E9 but changed to be right associative.

      E ::= T '+' E    // right recursion means right associative
          | T

      T ::= F '*' T    // right recursion means right associative
          | F

      F ::= '(' E ')'
          | 'a'

   We can rewrite the grammar this way.

      E ::= T [ '+' E ]   // right recursion means right associative
      T ::= F [ '*' T ]   // right recursion means right associative
      F ::= '(' E ')'
          | 'a'

   This parser does not "do" anything.
   It does not create a parse tree nor
   an abstract syntax tree. This parser
   is just a "recognizer" for the language.
   It will throw an error if the string cannot
   be parsed.

   Some example strings are:
      "(a)"
      "(a + a)"
      "(a + a) * a"
      "a * (a + a)"
      "(a + a) * (a + a)"
      "a * (a + (a * a))"
      "((a + a) * a) + a"
      "a * (a + a) * a"
*/
public class Parse_E9
{
   public static void recognizeE9(String stringInE9)
   {
      // Tokenize the input string.
      final Tokenizer tokens = new Tokenizer(stringInE9);
      // Parse the list of tokens.
      getE(tokens);
      // There shouldn't be any remaining tokens.
      if ( ! tokens.isEmpty() )
      {
         throw new ParseError("unexpected tokens, found " + tokens.peek());
      }
   }


   /**
      E ::= T [ '+' E ]
   */
   private static void getE(Tokenizer tokens)
   {
      getT(tokens);
      if ( ! tokens.isEmpty() && tokens.match("+") )
      {
         getE(tokens); // recursion
      }
   }


   /**
      T ::= F [ '*' T ]
   */
   private static void getT(Tokenizer tokens)
   {
      getF(tokens);
      if ( ! tokens.isEmpty() && tokens.match("*") )
      {
         getT(tokens); // recursion
      }
   }


   /**
      F ::= '(' E ')'
          | 'a'
   */
   private static void getF(Tokenizer tokens)
   {
      if ( tokens.match("(") )
      {
         getE(tokens);
         if ( ! tokens.match(")") )
         {
            throw new ParseError("expected \")\", found " + tokens.peek());
         }
      }
      else if ( ! tokens.match("a") )
      {
         throw new ParseError("expected \"a\", found " + tokens.peek());
      }
   }


   public static void main(String[] args)
   {
      String[] expressions = {
                              "(a)",
                              "(a + a)",
                              "(a + a) * a",
                              "a * (a + a)",
                              "(a + a) * (a + a)",
                              "a * (a + (a * a))",
                              "((a + a) * a) + a",
                              "a * (a + a) * a",
                              "a * ((a + a)) * a",
                              "a - (a * a) + a",
                              "a * (a + a) - a",
                              "a + (a * a)) + a",
                              "a + ((a * a) + a",
                              "(a + a + a) - a"
                             };

      for (final String str : expressions)
      {
         try
         {
            System.out.print( "Parsing: \"" + str + "\"" );
            recognizeE9( str );
            System.out.println( " ... Parsed" );
         }
         catch (TokenizeError e)
         {
            System.out.println( " ... Failed (" + e + ")" );
         }
         catch (ParseError e)
         {
            System.out.println( " ... Failed (" + e + ")" );
         }
      }
   }
}
